/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.PointCloud;
import renderer.scene.util.MeshMaker;
import renderer.models_L.turtlegraphics.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**

*/
public class InteractiveTurtleModels_R2 implements KeyListener,
                                                   ComponentListener
{
   private boolean perspective = true;

   private Scene scene;      // The event handlers need
   private final JFrame jf;  // access to these fields.
   private final FrameBufferPanel fbp;
   private int currentModel = 0;
   private double deltaX = 0.0;
   private double deltaY = 0.0;
   private double deltaZ = 0.0;
   private boolean displayTranslation = false;
   private Model savedModel = null; // used to hold a PointCloud model
   private int pointSize = 0;       // used by the point clouds

   private boolean takeScreenshot = false;
   private int screenshotNumber = 0;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public InteractiveTurtleModels_R2()
   {
      // Create the Scene object that we shall render
      scene = new Scene("InteractiveTurtleModels_R2");

      int N = 9;

      for (int i = 0; i <= 5; ++i)
      {
         final Model m = new Model("polygasket_"+N+"_"+i);
         new PolygasketTurtle(m, N, i, -1.0, 0.0, -2.0);
         scene.addPosition(new Position(m));
      }

      N = 7;

      for (int i = 0; i <= 6; ++i)
      {
         final Model m = new Model("polygasket_"+N+"_"+i);
         new PolygasketTurtle(m, N, i, -1.0, 0.0, -2.0);
         scene.addPosition(new Position(m));
      }

      for (int i = 0; i <= 8; ++i)
      {
         final Model m = new Model("pentagasket " + i);
         new PentagasketTurtle(m, i, -0.6, -0.6, -1.0);
         scene.addPosition(new Position(m));
      }

      for (int i = 0; i <= 8; ++i)
      {
         final Model m = new Model("sierpinski " + i);
         new SierpinskiTurtle(m, i, 4.0, -1.5, -2.0, -2.0);
         scene.addPosition(new Position(m));
      }

      for (int i = 0; i <= 10; ++i)
      {
         final Model m = new Model("sierpinski curve " + i);
         new SierpinskiCurveTurtle(m, i, 4.0, -1.5, -2.0, -2.0);
         scene.addPosition(new Position(m));
      }

      for (int i = 0; i <= 8; ++i)
      {
         final Model m = new Model("hilbert curve " + i);
         new HilbertCurveTurtle(m, i, 2.0, 0.1, -1.9, -2.0);
         scene.addPosition(new Position(m));
      }


      final Model ninjaModel = new Model("Ninja Turtle");
      new NinjaTurtle(ninjaModel, 180, 0.0, 0.0, -2.0);
      scene.addPosition(new Position(ninjaModel));

      final Model ninjaModel2 = new Model("Ninja Turtle");
      new NinjaTurtle(ninjaModel2, 360, 0.0, 0.0, -2.0);
      scene.addPosition(new Position(ninjaModel2));

      final Model spiralModel = new Model("Spiral Turtle");
      new SpiralTurtle(spiralModel, 270, -0.3, -0.5, -0.6);
      scene.addPosition(new Position(spiralModel));

      final Model spiralModel2 = new Model("Spiral Turtle");
      new SpiralTurtle(spiralModel2, 335, -0.3, -0.5, -0.6);
      scene.addPosition(new Position(spiralModel2));


      final Model turtleModel = new Model("Turtle Expeiment");
      final Turtle turtle = new Turtle(turtleModel, -2.0, 0.0, -3.0);

      final int angle = 144;
      final double length = 1.0;
      // draw a pentagram
      final int n = 5;
      for (int i = 1; i <= n; ++i)
      {
         turtle.forward(length);
         turtle.turn(angle);
      }
      for (int j = 1; j < 32; ++j)
      {
// https://www.clear.rice.edu/comp360/lectures/fall2008/TurtleFractalsL2New.pdf#page=3
         for (int i = 1; i <= 8; ++i)
         {
            turtle.forward(1);
            turtle.turn(60);
            turtle.resize(0.99);
            turtle.forward(1.0/4.0);
            turtle.turn(45);
            turtle.resize(0.99);
            turtle.forward(1.0/3.0);
            turtle.turn(-90);
            turtle.resize(0.99);
            turtle.forward(3.0/5.0);
            turtle.turn(30);
            turtle.resize(0.99);
         }
      }
/*
      for (int i = 1; i <= 8; ++i)
      {
         turtle.forward(1);
         turtle.turn(60);
         turtle.forward(1.0/4.0);
         turtle.turn(45);
         turtle.forward(1.0/3.0);
         turtle.turn(-90);
         turtle.forward(3.0/5.0);
         turtle.turn(30);
      }
*/
/*
      for (int j = 1; j < 32; ++j)
      {
         for (int i = 1; i <= 8; ++i)
         {
            turtle.forward(1);
            turtle.turn(60);
            turtle.forward(1.0/4.0);
            turtle.turn(45);
            turtle.forward(1.0/3.0);
            turtle.turn(-90);
            turtle.forward(3.0/5.0);
            turtle.turn(30);
         }
         turtle.resize(0.9);
      }
*/
      scene.addPosition(new Position(turtleModel));



      // Make the models invisible, except for the current model.
      for (Position p : scene.positionList)
      {
         p.visible = false;
      }
      currentModel = 0;
      scene.getPosition(currentModel).visible = true;

      Rasterize.doClipping = true;


      // Create a FrameBufferPanel that holds a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;
      fbp = new FrameBufferPanel(width, height);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 2 - Turtle Geometry");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      print_help_message();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println(scene.getPosition(currentModel).getModel());
      }
      else if ('d' == c)
      {
         scene.debug = ! scene.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('i' == c)
      {
         final int verts = scene.getPosition(currentModel).getModel().vertexList.size();
         final int lines = scene.getPosition(currentModel).getModel().primitiveList.size();
         System.out.printf("The current Model has %,d vertices and ", verts);
         System.out.printf("%,d line segments.\n", lines);
      }
      else if ('c' == c)
      {
         Rasterize.doClipping = ! Rasterize.doClipping;
         System.out.print("Clipping is turned ");
         System.out.println(Rasterize.doClipping ? "On" : "Off");
      }
      else if ('/' == c)
      {
         scene.getPosition(currentModel).visible = false;
         currentModel = (currentModel + 1) % scene.positionList.size();
         scene.getPosition(currentModel).visible = true;
         savedModel = null;
         pointSize = 0;
      }
      else if ('?' == c)
      {
         scene.getPosition(currentModel).visible = false;
         currentModel = currentModel - 1;
         if (currentModel < 0) currentModel = scene.positionList.size() - 1;
         scene.getPosition(currentModel).visible = true;
         savedModel = null;
         pointSize = 0;
      }
      else if ('p' == c)
      {
         perspective = ! perspective;
         final String p = perspective ? "perspective" : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('P' == c)
      {
         if (savedModel != null)
         {
            scene.getPosition(currentModel).setModel(savedModel);
            savedModel = null;
            ++pointSize;
         }
         else
         {
            final Model model = scene.getPosition(currentModel).getModel();
            savedModel = model;
            scene.getPosition(currentModel).setModel(
                                 PointCloud.make(model, pointSize));
         }
      }
      else if ('m' == c) // Display translation information.
      {
         displayTranslation = ! displayTranslation;
      }
      else if ('=' == c) // Reset the translation vector.
      {
         deltaX =  0;
         deltaY =  0;
         deltaZ =  0;
      }
      else if ('x' == c) // Translate ALL the models.
      {
         deltaX += -0.1;  // left
      }
      else if ('X' == c)
      {
         deltaX += +0.1;  // right
      }
      else if ('y' == c)
      {
         deltaY += -0.1;  // down
      }
      else if ('Y' == c)
      {
         deltaY += +0.1;  // up
      }
      else if ('z' == c)
      {
         deltaZ += -0.1;  // back
      }
      else if ('Z' == c)
      {
         deltaZ += +0.1;  // forward
      }
      else if ('+' == c)
      {
         takeScreenshot = true;
      }

      // Set up the camera's view volume.
      if (perspective)
      {
         scene = scene.changeCamera( Camera.projPerspective() );
      }
      else
      {
         scene = scene.changeCamera( Camera.projOrtho() );
      }

      scene.getPosition(currentModel).translate(deltaX, deltaY, deltaZ);

      if (displayTranslation && ('m'==c||'='==c
                               ||'x'==c||'y'==c||'z'==c
                               ||'X'==c||'Y'==c||'Z'==c))
      {
         System.out.printf("deltaX = % .8f, deltaY = % .8f, " +
                           "deltaZ = % .8f\n",
                            deltaX, deltaY, deltaZ);
      }

      // Render again.
      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      if (takeScreenshot)
      {
         fb.dumpFB2File(String.format("Screenshot%03d.png", screenshotNumber),
                        "png");
         ++screenshotNumber;
         takeScreenshot = false;
      }
      fbp.repaint();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        fbp.getTopLevelAncestor().getWidth(),
                        fbp.getTopLevelAncestor().getHeight(),
                        fbp.getWidth(), fbp.getHeight());

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final FrameBuffer fb = new FrameBuffer(w, h);
      fbp.setFrameBuffer(fb);
      Pipeline.render(scene, fb);
      fbp.repaint();
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off for the current model.");
      System.out.println("Use the '/' and '?' keys to cycle forwards and backwards through the models.");
      System.out.println("Use the 'i' key to get information about the current model.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the models along the x, y, z axes.");
      System.out.println("Use the 'm' key to toggle the display of translation information.");
      System.out.println("Use the '=' key to reset the model translation.");
      System.out.println("Use the 'c' key to toggle line clipping on and off.");
      System.out.println("Use the 'P' key to convert the current model to a point cloud.");
      System.out.println("Use the '+' key to save a \"screenshot\" of the framebuffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new InteractiveTurtleModels_R2()
      );
   }
}
